from .fan_control import FanControlBuilder as FanControlBuilder
from .fan_control import MachineLayout as MachineLayout
from .fan_control import ModuleConfig as ModuleConfig
from .fan_control import ModuleType as ModuleType
from .models import ModuleInfo as ModuleInfo
from .models import Quat as Quat
from .models import TrackedProbeData as TrackedProbeData
from .models import TrackingData as TrackingData
from .models import Vec3 as Vec3
from .models import WindProbeData as WindProbeData
from .windsuite_sdk import WindsuiteSDK as WindsuiteSDK



"""
Windsuite SDK main module.
"""

from __future__ import annotations

import asyncio
import math
import traceback
from collections.abc import Awaitable, Callable
from concurrent.futures import ThreadPoolExecutor
from typing import Any, Protocol

import requests
import socketio  # type: ignore[import-untyped]
from loguru import logger

from ._models_internal import TrackingDataDict, dict_to_dataclass
from .fan_control import FanControlBuilder, MachineLayout
from .models import FullTrackedWindProbeData, ModuleInfo, ModuleType, TrackedProbeData, TrackingData, Vec3

logger.remove()


class CallbackWindtrackData(Protocol):
    """
    This represents a callback function called whenever new windtrack data is received.

    It will pass the received data for the user to use.

    Args:
        data (dict[str, TrackingData]): Dictionary mapping object names (str) to their tracking data (TrackingData)

    """

    def __call__(self, data: dict[str, TrackingData]) -> None: ...


class CallbackWindprobeData(Protocol):
    """
    This represents a callback function called whenever new windprobe data is received.

    It will pass the received data for the user to use.

    Args:
        data (WindProbeData): The windprobe data received

    """

    def __call__(self, data: WindProbeData) -> None: ...


class CallbackTrackedProbeData(Protocol):
    """
    This represents a callback function called whenever new tracked windprobe data is received.

    It will pass the received data for the user to use.

    Args:
        data (TrackedProbeData): The tracked windprobe data received

    Note:
    - This callback is separate from CallbackWindprobeData to allow users to choose
    whether they want just the windprobe data or the full tracked data.

    - It also gives the user access to the best windprobe data if available depending on tracking (See TrackedProbeData comment)

    """

    def __call__(self, data: TrackedProbeData) -> None: ...


class CallbackModuleUpdate(Protocol):
    """
    This represents a callback function called whenever new module update data is received.

    It will pass the received data for the user to use.

    Args:
        data (dict[tuple[int, int], ModuleInfo]): Dictionary mapping module positions (row, col) to their ModuleInfo

    """

    def __call__(self, data: dict[tuple[int, int], ModuleInfo]) -> None: ...


class WindsuiteSDK:
    def __init__(
        self,
        base_url: str,
        port_rest_api: int = 8000,
        sio_port: int = 8001,
        sio_port_rt: int = 8002,
    ) -> None:
        self.port_rest_api = port_rest_api
        self.sio_port = sio_port
        self.sio_port_rt = sio_port_rt

        self.rest_base_url = base_url.rstrip("/") + f":{self.port_rest_api}" + "/api"
        self.sio_base_url = base_url.rstrip("/") + f":{self.sio_port}"
        self.sio_rt_base_url = base_url.rstrip("/") + f":{self.sio_port_rt}"

        self.__callback_windprobe: CallbackWindprobeData | None = None
        self.__callback_tracked_probe_data: CallbackTrackedProbeData | None = None
        self.__callback_windtrack: CallbackWindtrackData | None = None
        self.__callback_module_update: CallbackModuleUpdate | None = None

        self._executor = ThreadPoolExecutor(max_workers=2, thread_name_prefix="windsuite")
        self._stop_events: list[asyncio.Event] = []
        self._loops: list[asyncio.AbstractEventLoop] = []

        self.__sio: socketio.AsyncClient
        self.__sio_rt: socketio.AsyncClient

        self.current_layout = MachineLayout.from_api(self.rest_base_url)
        self.fan_controller = FanControlBuilder(self.current_layout)

    def __start_async_thread(
        self,
        async_routine: Callable[[asyncio.Event], Awaitable[None]],
        thread_name: str,
    ) -> None:
        """
        Helper to start an async routine in a background thread using ThreadPoolExecutor.

        Args:
            async_routine: The async coroutine function to run (receives a stop_event)
            thread_name: Name for the thread

        """

        def run_async() -> None:
            loop: asyncio.AbstractEventLoop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop=loop)
            stop_event = asyncio.Event()

            # Store the loop and event for cleanup
            self._loops.append(loop)
            self._stop_events.append(stop_event)

            try:
                loop.run_until_complete(async_routine(stop_event))
            finally:
                loop.close()

        self._executor.submit(run_async)

    def start_communication(self) -> None:
        """
        Called to start real-time communication with the Windsuite server

        ---

        When calling this:
        - Windprobe data will be received and passed to the registered windprobe callback
        - Windtrack data will be received and passed to the registered windtrack callback
        - Module updates will be received and passed to the registered module update callback
        - Facility updates (layout changes) will be handled automatically
        """
        self.__start_async_thread(self.__socketio_client_routine, "socketio_client_thread")
        self.__start_async_thread(self.__socketio_rt_client_routine, "socketio_rt_client_thread")

    def cleanup(self) -> None:
        """
        Stops all communication threads and disconnects from socketio servers.
        """
        logger.info("Starting cleanup...")

        for loop, stop_event in zip(self._loops, self._stop_events, strict=True):
            loop.call_soon_threadsafe(stop_event.set)

        logger.info("Shutting down thread pool...")
        self._executor.shutdown(wait=True, cancel_futures=False)

        logger.info("Cleanup complete.")

    def zero_windprobe(self) -> bool:
        """
        Trigger the zeroing of the windprobe
        """
        try:
            response = requests.post(f"{self.rest_base_url}/windprobe/zero", timeout=5)
            response.raise_for_status()
            return True
        except requests.RequestException as e:
            logger.error(f"Error zeroing windprobe: {e}")
            return False

    def set_psu(self, state: bool) -> bool:
        """
        Set the PSU of all modules to ON or OFF.
        """
        try:
            response = requests.patch(
                f"{self.rest_base_url}/windcontrol/modules/psu?psu_state_to_set={str(state).lower()}", timeout=5
            )
            response.raise_for_status()
            return True
        except requests.RequestException as e:
            logger.error(f"Error setting PSU state: {e}")
            return False

    def __replace_sentinel_with_nan(self, vec: Vec3) -> None:
        """Replace -999 sentinel values with nan in a Vec3."""
        if vec.x == -999:
            vec.x = math.nan
        if vec.y == -999:
            vec.y = math.nan
        if vec.z == -999:
            vec.z = math.nan

    async def __handle_probe_data(self, data: dict[str, Any]) -> None:
        """Handle incoming probe data."""
        try:
            full_tracked_probe_data = dict_to_dataclass(FullTrackedWindProbeData, data)

            # ! WINDPROBE DATA
            # Replace -999 sentinel values with nan in wind velocity fields
            windprobe = full_tracked_probe_data.windprobe_data
            self.__replace_sentinel_with_nan(windprobe.wind_velocity_mps_probe_ref)
            self.__replace_sentinel_with_nan(windprobe.wind_velocity_mps_windshaper_ref)
            self.__replace_sentinel_with_nan(windprobe.wind_velocity_mps_windshaper_ref_corrected)

            if self.__callback_windprobe:
                self.__callback_windprobe(windprobe)

            # ! TRACKED PROBE DATA
            # ! UGLY STUFF, NEEDS REFACTORING ASAP
            if full_tracked_probe_data.curated_tracked_probe_data is not None:
                tracked_probe_data = TrackedProbeData(
                    timestamp_s=full_tracked_probe_data.windprobe_data.timestamp_s,
                    wind_velocity_mps=full_tracked_probe_data.curated_tracked_probe_data.best_windprobe_data.best_wind_velocity_mps,
                    position_data_mm=full_tracked_probe_data.curated_tracked_probe_data.probe_position_mm_windshaper_ref,
                    temperature_celcius=full_tracked_probe_data.curated_tracked_probe_data.best_windprobe_data.temperature_celcius,
                    atmospheric_pressure_hpascal=full_tracked_probe_data.curated_tracked_probe_data.best_windprobe_data.atmospheric_pressure_hpascal,
                    static_pressure_pascal=full_tracked_probe_data.curated_tracked_probe_data.best_windprobe_data.static_pressure_pascal,
                )
                if self.__callback_tracked_probe_data:
                    self.__callback_tracked_probe_data(tracked_probe_data)
            # else:
            # print("No best probe data available for this tracked probe data.")

        except (KeyError, TypeError, ValueError) as e:
            logger.error(f"Error parsing probe data: {e}")
            logger.debug(f"Received data keys: {list(data.keys())}")

    async def __handle_tracking_data(self, data: dict[str, Any]) -> None:
        """Handle incoming tracking data."""
        try:
            windtrack_dict = dict_to_dataclass(TrackingDataDict, data)
            if self.__callback_windtrack:
                self.__callback_windtrack(windtrack_dict.data)
        except (KeyError, TypeError, ValueError) as e:
            logger.error(f"Error parsing tracking data: {e}")
            logger.debug(f"Received data keys: {list(data.keys())}")

    async def __handle_module_update(self, data: dict[str, Any]) -> None:
        """Handle incoming module update data."""
        try:
            modules_info: dict[tuple[int, int], ModuleInfo] = {}
            updated_modules = data.get("updated_modules", {})
            for mac, module_data in updated_modules.items():
                # Look up row/col from the layout using MAC address
                position = self.current_layout.get_module_by_mac(mac)
                if position is None:
                    logger.warning(f"Received update for unknown module MAC: {mac}")
                    continue
                row, col = position

                # ! 1 based indexing
                row += 1
                col += 1

                pwms_data: dict[str, list[float]] = module_data.get("pwms", {})
                rpms_data: dict[str, list[float]] = module_data.get("rpms", {})

                # Convert PWM values from 0-1000 range to 0-100 range
                target_pwm: list[list[float]] = [[value / 10.0 for value in pwms_data[key]] for key in sorted(pwms_data.keys())]
                if not target_pwm:
                    target_pwm = [[]]

                # Convert PWM values from 0-1000 range to 0-100 range
                current_pwm: list[list[float]] = [[value / 10.0 for value in pwms_data[key]] for key in sorted(pwms_data.keys())]
                if not current_pwm:
                    current_pwm = [[]]

                # Build current_rpm as list of lists (one inner list per layer)
                current_rpm: list[list[float]] = [rpms_data[key] for key in sorted(rpms_data.keys())]
                if not current_rpm:
                    current_rpm = [[]]

                module_type_str = module_data["type"]
                if module_type_str in {e.value for e in ModuleType}:
                    module_type_str = ModuleType(module_type_str).value
                else:
                    logger.warning("Unknown module type received: {}", module_type_str)

                module_info = ModuleInfo(
                    row=row,
                    col=col,
                    mac=module_data["mac"],
                    ip=module_data["ip"],
                    type=module_type_str,
                    lifepoints=module_data.get("lifepoints"),
                    target_pwm=target_pwm,
                    current_pwm=current_pwm,
                    current_rpm=current_rpm,
                    target_psu_state=bool(module_data.get("psu_command")),
                    current_psu_state=bool(module_data.get("is_psu")),
                    is_connected=bool(module_data.get("is_connected")),
                    # ! Optional ESC and PSU data
                    esc_rpm=module_data.get("esc_rpm"),
                    applied_pwm_us=module_data.get("applied_pwm_us"),
                    esc_voltage_mv=module_data.get("esc_voltage_mv"),
                    esc_bus_current_ma=module_data.get("esc_bus_current_ma"),
                    esc_motor_current_ma=module_data.get("esc_motor_current_ma"),
                    esc_index=module_data.get("esc_index"),
                    esc_temp_mos_celcius=module_data.get("esc_temp_mos_celcius"),
                    esc_temp_capacitor_celcius=module_data.get("esc_temp_capacitor_celcius"),
                    esc_temp_mcu_celcius=module_data.get("esc_temp_mcu_celcius"),
                    esc_temp_motor_celcius=module_data.get("esc_temp_motor_celcius"),
                    esc_running_error=module_data.get("esc_running_error"),
                    esc_uptime=module_data.get("esc_uptime"),
                    applied_psu=module_data.get("applied_psu"),
                    psu_v_in=module_data.get("psu_v_in"),
                    psu_i_in=module_data.get("psu_i_in"),
                    psu_p_in=module_data.get("psu_p_in"),
                    psu_v_out=module_data.get("psu_v_out"),
                    psu_i_out=module_data.get("psu_i_out"),
                    psu_p_out=module_data.get("psu_p_out"),
                    psu_temp1=module_data.get("psu_temp1"),
                    psu_temp2=module_data.get("psu_temp2"),
                    psu_temp3=module_data.get("psu_temp3"),
                )
                modules_info[(row, col)] = module_info

            if self.__callback_module_update:
                self.__callback_module_update(modules_info)
            logger.debug("Parsed {} module(s): {}", len(modules_info), list(modules_info.keys()))
        except (KeyError, TypeError, ValueError) as e:
            logger.error(f"Error parsing module update data: {e} - {traceback.format_exc()}")
            logger.debug("Received data keys: {}", list(data.keys()))

    async def __handle_facility_update(self, data: dict[str, Any]) -> None:
        """Handle incoming facility update data (layout changes)."""
        try:
            windcontrol_layout = data.get("current_windcontrol_layout")
            if windcontrol_layout is None:
                logger.warning("Received facility_update without current_windcontrol_layout")
                return

            # Preserve the API URL from the current layout
            api_url = self.current_layout.api_url

            # Update the layout
            self.current_layout = MachineLayout.from_dict(windcontrol_layout, api_url=api_url)
            self.fan_controller = FanControlBuilder(self.current_layout)

            logger.info(
                f"Layout updated: {self.current_layout.nb_rows}x{self.current_layout.nb_columns} with {len(self.current_layout.modules)} modules"
            )
        except (KeyError, TypeError, ValueError) as e:
            logger.error(f"Error parsing facility update data: {e} - {traceback.format_exc()}")
            logger.debug("Received data keys: {}", list(data.keys()))

    async def __socketio_client_routine(self, stop_event: asyncio.Event) -> None:
        """
        Base socketio client for general events (windprobe data)
        """
        self.__sio: socketio.AsyncClient = socketio.AsyncClient(
            reconnection=False,
            request_timeout=5,
            reconnection_delay=1,
            reconnection_delay_max=5,
        )
        # ? Ignore rationales: sio is badly typed

        @self.__sio.event  # pyright: ignore[reportUnknownMemberType]
        async def connect() -> None:  # pyright: ignore[reportUnusedFunction]
            logger.info("Connected to SocketIO server.")

        @self.__sio.event  # pyright: ignore[reportUnknownMemberType]
        async def disconnect() -> None:  # pyright: ignore[reportUnusedFunction]
            logger.info("Disconnected from SocketIO server.")

        @self.__sio.event  # pyright: ignore[reportUnknownMemberType]
        async def probe_data(data: dict[str, Any]) -> None:  # pyright: ignore[reportUnusedFunction]
            await self.__handle_probe_data(data)

        @self.__sio.event  # pyright: ignore[reportUnknownMemberType]
        async def module_update(data: dict[str, Any]) -> None:  # pyright: ignore[reportUnusedFunction]
            await self.__handle_module_update(data)

        @self.__sio.event  # pyright: ignore[reportUnknownMemberType]
        async def facility_update(data: dict[str, Any]) -> None:  # pyright: ignore[reportUnusedFunction]
            await self.__handle_facility_update(data)

        try:
            await self.__sio.connect(  # pyright: ignore[reportUnknownMemberType]
                self.sio_base_url,
                transports=["websocket"],
                wait=True,
                wait_timeout=1,
            )
        except socketio.exceptions.ConnectionError as err:  # type: ignore[UnknownMemberType], Rationale : type hints of socketio are not strict enough
            logger.error(
                f"Socket IO connection error: {err}\nHINT:\n\t- Verify that the server is running\n\t- The URL is correct\n\t- You're on the same network as the server"
            )
            return

        await stop_event.wait()

        logger.info("Shutting down SIO client...")
        await self.__sio.disconnect()  # pyright: ignore[reportUnknownMemberType]
        logger.info("SIO client disconnected.")

    async def __socketio_rt_client_routine(self, stop_event: asyncio.Event) -> None:
        """
        Real-time socketio client
        """
        self.__sio_rt: socketio.AsyncClient = socketio.AsyncClient(
            reconnection=False,
            request_timeout=5,
            reconnection_delay=1,
            reconnection_delay_max=5,
        )
        # ? Ignore rationales: sio is badly typed

        @self.__sio_rt.event  # pyright: ignore[reportUnknownMemberType]
        async def connect() -> None:  # pyright: ignore[reportUnusedFunction]
            logger.info("Connected to SocketIO RT server.")

        @self.__sio_rt.event  # pyright: ignore[reportUnknownMemberType]
        async def disconnect() -> None:  # pyright: ignore[reportUnusedFunction]
            logger.info("Disconnected from SocketIO RT server.")

        @self.__sio_rt.event  # pyright: ignore[reportUnknownMemberType]
        async def tracking_data(data: dict[str, Any]) -> None:  # pyright: ignore[reportUnusedFunction]
            await self.__handle_tracking_data(data)

        try:
            await self.__sio_rt.connect(  # pyright: ignore[reportUnknownMemberType]
                self.sio_rt_base_url,
                transports=["websocket"],
                wait=True,
                wait_timeout=1,
            )
        except socketio.exceptions.ConnectionError as err:  # type: ignore[UnknownMemberType], Rationale : type hints of socketio are not strict enough
            logger.error(
                f"Socket IO RT connection error: {err}\nHINT:\n\t- Verify that the server is running\n\t- The URL is correct\n\t- You're on the same network as the server"
            )
            return

        await stop_event.wait()

        logger.info("Shutting down SIO RT client...")
        await self.__sio_rt.disconnect()  # pyright: ignore[reportUnknownMemberType]
        logger.info("SIO RT client disconnected.")

    def register_tracking_callback(
        self,
        callback: CallbackWindtrackData,
    ) -> None:
        """
        Register a callback for windtrack data.

        Args:
            callback (CallbackWindtrackData): The function to call when windtrack data is received.
                                         Receives a dict mapping object names to TrackingData.

        """
        if not callable(callback):
            raise TypeError("Callback must be a callable function.")

        self.__callback_windtrack = callback

    def register_windprobe_callback(
        self,
        callback: CallbackWindprobeData,
    ) -> None:
        """
        Register a callback for windprobe data.

        Args:
            callback (CallbackWindprobeData): The function to call when windprobe data is received.

        """
        if not callable(callback):
            raise TypeError("Callback must be a callable function.")

        self.__callback_windprobe = callback

    def register_tracked_probe_callback(
        self,
        callback: CallbackTrackedProbeData,
    ) -> None:
        """
        Register a callback for tracked windprobe data.

        Args:
            callback (CallbackTrackedProbeData): The function to call when tracked windprobe data is received.

        """
        if not callable(callback):
            raise TypeError("Callback must be a callable function.")

        self.__callback_tracked_probe_data = callback

    def register_module_update_callback(
        self,
        callback: CallbackModuleUpdate,
    ) -> None:
        """
        Register a callback for module update data.

        Args:
            callback (CallbackModuleUpdate): The function to call when module update data is received.

        """
        if not callable(callback):
            raise TypeError("Callback must be a callable function.")

        self.__callback_module_update = callback



"""Data models for the Windsuite SDK."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import ClassVar

# ! -------------
# ! USER FACING
# ! -------------


@dataclass
class Vec3:
    """3D vector with x, y, z components."""

    x: float
    y: float
    z: float


@dataclass
class Quat:
    """Quaternion with w, x, y, z components."""

    w: float
    x: float
    y: float
    z: float


@dataclass
class WindProbeData:
    """
    Represents all data from a wind probe.

    This includes:
    - Raw data
    - Data corrected for windshaper orientation
    - Data corrected for windshaper and probe movement / orientation

    Notes:
    Depending on the tracking quality, some data might contain NaN values.

    If you want the best possible, you either need to check the tracking data yourself
    or use the BestWindProbeData callback that will give you the best possible Wind Data and Tracking Data directly.

    All wind velocity data is in meters per second (mps).

    """

    timestamp_s: float
    wind_velocity_mps_probe_ref: Vec3  # ! Raw data
    wind_velocity_mps_windshaper_ref: Vec3  # ! Corrected in windshaper reference frame
    wind_velocity_mps_windshaper_ref_corrected: Vec3  # ! Corrected for probe movement / orientation in windshaper reference frame

    temperature_celcius: float
    atmospheric_pressure_hpascal: float
    static_pressure_pascal: float

    conversion_status: int
    data_is_valid: bool


@dataclass
class TrackingData:
    """
    Represents tracking data for a single object.

    - Position in meters in the world reference frame,
    - Orientation as a quaternion in the world reference frame,

    ! Special case for the windshaper itself that will have the same as it's his reference frame
    - Position in meters in the windshaper reference frame,
    - Orientation as a quaternion in the windshaper reference frame,

    - Velocity in meters per second,
    - Angular velocity in degrees per second.
    """

    timestamp: float

    position_meters_world_ref: Vec3
    rotation_world_ref: Quat

    position_meters_windshaper_ref: Vec3
    rotation_windshaper_ref: Quat

    velocity_mps_world_ref: Vec3
    angular_velocity_degps_world_ref: Vec3

    velocity_mps_windshaper_ref: Vec3
    angular_velocity_degps_windshaper_ref: Vec3

    is_tracked: bool = False


@dataclass
class TrackedProbeData:
    """
    Represents the best Wind data associated with the current position of the probe.

    Notes:
    The best data is chosen based on the tracking quality to provide the most accurate wind data possible.
    If there is no tracking, the raw wind data is used.
    If there is unstable tracking data, the orientation-corrected wind data is used.
    If there is stable tracking data, the movement / orientation corrected wind data is used.

    The position is given in the windshaper reference frame (if possible) in millimeters.

    """

    timestamp_s: float

    wind_velocity_mps: Vec3
    position_data_mm: Vec3

    temperature_celcius: float
    atmospheric_pressure_hpascal: float
    static_pressure_pascal: float


@dataclass
class ModuleInfo:
    """
    Represents the current state of a Windshaper Module.

    Notes:
    It's possible to receive rpms and pwms in different formats depending on the Module type.

    For simplicity of use, the following conventions are used:
        Every received information is a list of lists, where each inner list represents a layer and each of these list contains the values for each fan in that layer.

        e.g.
            ! Module type 2420
            pwms = [[25.0]] -> Module with 1 layer and 1 fan with 25% pwm
            rpms = [[3445]] -> Module with 1 layer and 1 fan with 3445 rpm

            ! Module type 0816 / 0812
            pwms = [
                        [10, 10, 20, 20, 10, 10, 20, 20, 30], # ! UPSTREAM LAYER
                        [40, 40, 50, 50, 40, 40, 50, 50, 30], # ! DOWNSTREAM LAYER
                    ] -> Module with 2 layers and 9 fans per layer
            rpms = [
                        [1200, 1250, 1300, 1280, 1190, 1230, 1310, 1290, 1150], # ! UPSTREAM LAYER
                        [2200, 2250, 2300, 2280, 2190, 2230, 2310, 2290, 2150], # ! DOWNSTREAM LAYER
                    ] -> Module with 2 layers and 9 fans per layer

    """

    INDEX_DOWNSTREAM: ClassVar[int] = 0
    INDEX_UPSTREAM: ClassVar[int] = 1

    row: int
    col: int
    mac: str
    ip: str
    type: str

    lifepoints: int

    target_pwm: list[list[float]]
    current_pwm: list[list[float]]

    current_rpm: list[list[float]]

    target_psu_state: bool
    current_psu_state: bool

    is_connected: bool = False

    esc_rpm: int | None = None
    applied_pwm_us: int | None = None
    esc_voltage_mv: int | None = None
    esc_bus_current_ma: int | None = None
    esc_motor_current_ma: int | None = None
    esc_index: int | None = None
    esc_temp_mos_celcius: int | None = None
    esc_temp_capacitor_celcius: int | None = None
    esc_temp_mcu_celcius: int | None = None
    esc_temp_motor_celcius: int | None = None
    esc_running_error: int | None = None
    esc_uptime: int | None = None
    applied_psu: int | None = None
    psu_v_in: float | None = None
    psu_i_in: float | None = None
    psu_p_in: float | None = None
    psu_v_out: float | None = None
    psu_i_out: float | None = None
    psu_p_out: float | None = None
    psu_temp1: float | None = None
    psu_temp2: float | None = None
    psu_temp3: float | None = None


# ! -------------
# ! INTERNAL
# ! -------------


@dataclass
class FullTrackedWindProbeData:
    """
    Represents windprobe data associated with the current position of the probe.
    """

    probe_tracking_data: TrackingData
    windprobe_data: WindProbeData
    curated_tracked_probe_data: BestFullTrackedWindProbeData | None = None


@dataclass
class BestWindProbeData:
    """
    Represents curated data from a wind probe.
    The velocity is the most accurate depending on the tracking.
    """

    timestamp_s: float
    best_wind_velocity_mps: Vec3
    temperature_celcius: float
    atmospheric_pressure_hpascal: float
    static_pressure_pascal: float
    conversion_status: int


@dataclass
class BestFullTrackedWindProbeData:
    """
    Represents curated windprobe data associated with the current position of the probe.
    """

    probe_position_mm_windshaper_ref: Vec3
    best_windprobe_data: BestWindProbeData


class ModuleType(str, Enum):
    """
    Enum representing the type of a Windshaper Module.
    """

    MODULE_0812 = "0812"
    MODULE_0816 = "0816"
    MODULE_2420 = "2420"



"""Internal utilities for parsing data into dataclasses."""

from __future__ import annotations

import types
from dataclasses import dataclass, fields, is_dataclass
from enum import Enum
from typing import Any, TypeVar, Union, cast, get_args, get_origin, get_type_hints

T = TypeVar("T")

# Import at runtime for get_type_hints to resolve forward references
from .models import TrackingData


@dataclass
class TrackingDataDict:
    """Internal wrapper for tracking data dictionary."""

    data: dict[str, TrackingData]


class WindProbeState(int, Enum):
    """
    Enum for the state of the WindProbe
    """

    LOOKING_FOR_PROBE_ON_USB = 0
    HANDSHAKING_PROBE = 1
    PROBE_CONNECTED = 2
    PROBE_ZEROING_SENSORS = 3
    PROBE_FULLY_READY = 4


def dict_to_dataclass(cls: type[T], data: dict[str, Any]) -> T:
    """
    Convert a dictionary to a dataclass instance, recursively handling nested dataclasses.

    Args:
        cls: The dataclass type to create
        data: Dictionary containing the data

    Returns:
        An instance of the dataclass

    """
    if not is_dataclass(cls):
        raise TypeError(f"{cls} is not a dataclass")

    type_hints = get_type_hints(cls)

    field_values: dict[str, Any] = {}

    for field in fields(cls):
        if field.name not in data:
            # ! Skip missing fields if they have defaults
            continue

        value = data[field.name]
        field_type = type_hints.get(field.name, field.type)

        if value is None:
            field_values[field.name] = None
            continue

        origin = get_origin(field_type)
        if origin is dict:
            args = get_args(field_type)
            if len(args) == 2 and is_dataclass(args[1]):
                value_type = cast("type[Any]", args[1])
                field_values[field.name] = {k: dict_to_dataclass(value_type, v) for k, v in value.items()}
            else:
                field_values[field.name] = value

        elif is_dataclass(field_type) and isinstance(value, dict):
            field_values[field.name] = dict_to_dataclass(cast("type[Any]", field_type), value)

        # Handle Union types (e.g., X | None or Optional[X])
        elif origin in (Union, types.UnionType) and isinstance(value, dict):
            args = get_args(field_type)
            # Find the dataclass type in the union (skip NoneType)
            dataclass_type = next((arg for arg in args if is_dataclass(arg)), None)
            if dataclass_type is not None:
                field_values[field.name] = dict_to_dataclass(cast("type[Any]", dataclass_type), cast("dict[str, Any]", value))
            else:
                field_values[field.name] = value

        else:
            field_values[field.name] = value

    return cls(**field_values)



from __future__ import annotations

import time
from collections.abc import Callable
from dataclasses import dataclass
from enum import Enum
from typing import Any

import requests
from loguru import logger


class ModuleType(Enum):
    MODULE_2420 = "2420"
    MODULE_0816 = "0816"


@dataclass
class ModuleConfig:
    module_type: ModuleType
    row: int
    col: int
    mac: str | None = None

    @property
    def num_layers(self) -> int:
        return 2 if self.module_type == ModuleType.MODULE_0816 else 1

    @property
    def fans_per_layer(self) -> int:
        return 9 if self.module_type == ModuleType.MODULE_0816 else 1

    @property
    def fan_grid_rows(self) -> int:
        """Number of fan rows in this module type (3 for 0816, 1 for 2420)."""
        return 3 if self.module_type == ModuleType.MODULE_0816 else 1

    @property
    def fan_grid_cols(self) -> int:
        """Number of fan columns in this module type (3 for 0816, 1 for 2420)."""
        return 3 if self.module_type == ModuleType.MODULE_0816 else 1


class MachineLayout:
    def __init__(self, nb_rows: int, nb_columns: int, api_url: str | None = None):
        self.nb_rows = nb_rows
        self.nb_columns = nb_columns
        self.api_url = api_url
        self.modules: dict[tuple[int, int], ModuleConfig] = {}
        self.fan_states: dict[tuple[int, int, int, int], float] = {}
        self.modified_modules: set[tuple[int, int]] = set()
        self.flat_index_map: dict[tuple[int, int], int] = {}

    def add_module(
        self, row: int, col: int, module_type: ModuleType, flat_index: int | None = None, mac: str | None = None
    ) -> None:
        if row < 0 or row >= self.nb_rows:
            raise ValueError(f"Row {row} out of bounds [0, {self.nb_rows})")
        if col < 0 or col >= self.nb_columns:
            raise ValueError(f"Column {col} out of bounds [0, {self.nb_columns})")

        config = ModuleConfig(module_type, row, col, mac=mac)
        self.modules[(row, col)] = config

        if flat_index is not None:
            self.flat_index_map[(row, col)] = flat_index
        else:
            self.flat_index_map[(row, col)] = row * self.nb_columns + col

        for layer in range(config.num_layers):
            for fan_idx in range(config.fans_per_layer):
                self.fan_states[(row, col, layer, fan_idx)] = 0.0

    def get_module(self, row: int, col: int) -> ModuleConfig | None:
        return self.modules.get((row, col))

    def get_module_by_mac(self, mac: str) -> tuple[int, int] | None:
        """Get (row, col) for a module by its MAC address."""
        for (row, col), config in self.modules.items():
            if config.mac == mac:
                return (row, col)
        return None

    def set_fan_pwm(self, row: int, col: int, layer: int, fan_index: int, pwm: float) -> None:
        pwm = max(0.0, min(100.0, pwm))
        self.fan_states[(row, col, layer, fan_index)] = pwm
        self.modified_modules.add((row, col))

    def get_fan_pwm(self, row: int, col: int, layer: int, fan_index: int) -> float:
        return self.fan_states.get((row, col, layer, fan_index), 0.0)

    @classmethod
    def from_dict(cls, windcontrol_layout: dict[str, Any], api_url: str | None = None) -> MachineLayout:
        """Create a MachineLayout from a windcontrol_layout dictionary."""
        size_x = windcontrol_layout["size_x"]
        size_y = windcontrol_layout["size_y"]
        module_grid = windcontrol_layout["module_grid"]

        layout = cls(nb_rows=size_y, nb_columns=size_x, api_url=api_url)

        for flat_idex_str, module_data in module_grid.items():
            flat_index = int(flat_idex_str)
            row = module_data["grid_index_y"]
            col = module_data["grid_index_x"]
            module_type_str = module_data["module_type"]
            mac = module_data.get("mac")

            if module_type_str == "2420":
                module_type = ModuleType.MODULE_2420
            elif module_type_str == "0816":
                module_type = ModuleType.MODULE_0816
            else:
                continue

            layout.add_module(row, col, module_type, flat_index=flat_index, mac=mac)

        return layout

    @classmethod
    def from_api(cls, api_url: str) -> MachineLayout:
        try:
            url = f"{api_url.rstrip('/')}/windcontrol/layouts/"
            response = requests.get(url, timeout=5)
            response.raise_for_status()

            data = response.json()
            windcontrol_layout = data["current_windcontrol_layout"]
        except requests.exceptions.ConnectionError as e:
            logger.error(f"HTTP connection error while fetching layout: {e}\nCheck API URL and network connection.")
            raise
        except Exception as e:
            logger.error(f"HTTP connection error while fetching layout: {e}\nCheck API URL and network connection.")
        return cls.from_dict(windcontrol_layout, api_url=api_url)


class FanControlBuilder:
    COOLDOWN_SECONDS: float = 1.0

    def __init__(self, layout: MachineLayout):
        self.layout = layout
        self._last_error_time: float = 0.0
        self._reset_selection()

    def _reset_selection(self) -> None:
        self._selected_rows: list[int] | None = None
        self._selected_cols: list[int] | None = None
        self._selected_layer: int | None = None
        self._selected_fans: list[int] | None = None
        self._selected_parity: int | None = None  # 0 = even, 1 = odd

    def _convert_index(self, index: int, max_val: int) -> int:
        """Convert potentially negative 1-based index to positive 1-based index."""
        if index < 0:
            return max_val + index + 1
        return index

    def _validate_index(self, index: int, max_val: int, name: str) -> None:
        """Validate a single 1-based index (supports negative indexing)."""
        if index == 0:
            raise ValueError(f"{name} index cannot be zero")
        if index > max_val or index < -max_val:
            raise ValueError(f"{name} index {index} out of bounds for layout with {max_val} {name.lower()}s")

    def _validate_indices(self, indices: list[int], max_val: int, name: str) -> None:
        """Validate a list of 1-based indices (supports negative indexing)."""
        if not indices:
            raise ValueError(f"'{name}' list cannot be empty")
        if any(i == 0 for i in indices):
            raise ValueError(f"'{name}' list cannot contain zero")
        if any(i > max_val or i < -max_val for i in indices):
            raise ValueError(f"'{name}' list contains out of bounds indices for layout with {max_val} {name.lower()}s")

    def _validate_range(self, from_val: int, to_val: int, max_val: int, name: str) -> None:
        """Validate from/to range parameters (supports negative indexing)."""
        if any(v == 0 for v in (from_val, to_val)):
            raise ValueError(f"'from_{name}' and 'to_{name}' cannot be zero")
        if any(v > max_val or v < -max_val for v in (from_val, to_val)):
            raise ValueError(f"'from_{name}' or 'to_{name}' out of bounds for layout with {max_val} {name}s")

    def rows(
        self,
        rows: list[int] | None = None,
        from_row: int | None = None,
        to_row: int | None = None,
    ) -> FanControlBuilder:
        """
        Select specific rows or a range of rows.

        Args:
            rows (list[int] | None): List of 1-based row indices to select.
            from_row (int | None): 1-based starting row index for range selection.
            to_row (int | None): 1-based ending row index for range selection.

        Notes:
            - Negative indices are supported (e.g., -1 refers to the last row).
            - Either 'rows' or 'from_row'/'to_row' should be provided, not both.
            - If using range selection, both 'from_row' and 'to_row' must be specified.

        """
        if rows is not None and (from_row is not None or to_row is not None):
            raise ValueError("Cannot use both 'rows' and 'from_row/to_row' parameters")

        nb_rows = self.layout.nb_rows
        if rows is not None:
            self._validate_indices(rows, nb_rows, "rows")
            self._selected_rows = [self._convert_index(r, nb_rows) - 1 for r in rows]
        elif from_row is not None and to_row is not None:
            self._validate_range(from_row, to_row, nb_rows, "row")
            from_conv = self._convert_index(from_row, nb_rows)
            to_conv = self._convert_index(to_row, nb_rows)
            if to_conv < from_conv:
                raise ValueError("'to_row' must be greater than or equal to 'from_row' after index conversion")
            self._selected_rows = list(range(from_conv - 1, to_conv))
        elif from_row is not None or to_row is not None:
            raise ValueError("Both 'from_row' and 'to_row' must be specified for range")

        return self

    def row(self, row: int) -> FanControlBuilder:
        """
        Select a specific single row. (use rows() to select multiple rows)

        Args:
            row (int): 1-based row index to select.

        Notes:
            - Negative indices are supported (e.g., -1 refers to the last row).

        """
        nb_rows = self.layout.nb_rows
        self._validate_index(row, nb_rows, "Row")
        self._selected_rows = [self._convert_index(row, nb_rows) - 1]
        return self

    def columns(
        self,
        columns: list[int] | None = None,
        from_col: int | None = None,
        to_col: int | None = None,
    ) -> FanControlBuilder:
        """
        Select specific columns or a range of columns.

        Args:
            columns (list[int] | None): List of 1-based column indices to select.
            from_col (int | None): 1-based starting column index for range selection.
            to_col (int | None): 1-based ending column index for range selection.

        Notes:
            - Negative indices are supported (e.g., -1 refers to the last column).
            - Either 'columns' or 'from_col'/'to_col' should be provided, not both.
            - If using range selection, both 'from_col' and 'to_col' must be specified.

        """
        if columns is not None and (from_col is not None or to_col is not None):
            raise ValueError("Cannot use both 'columns' and 'from_col/to_col' parameters")

        nb_columns = self.layout.nb_columns
        if columns is not None:
            self._validate_indices(columns, nb_columns, "columns")
            self._selected_cols = [self._convert_index(c, nb_columns) - 1 for c in columns]
        elif from_col is not None and to_col is not None:
            self._validate_range(from_col, to_col, nb_columns, "col")
            from_conv = self._convert_index(from_col, nb_columns)
            to_conv = self._convert_index(to_col, nb_columns)
            if to_conv < from_conv:
                raise ValueError("'to_col' must be greater than or equal to 'from_col' after index conversion")
            self._selected_cols = list(range(from_conv - 1, to_conv))
        elif from_col is not None or to_col is not None:
            raise ValueError("Both 'from_col' and 'to_col' must be specified for range")

        return self

    def column(self, column: int) -> FanControlBuilder:
        nb_columns = self.layout.nb_columns
        self._validate_index(column, nb_columns, "Column")
        self._selected_cols = [self._convert_index(column, nb_columns) - 1]
        return self

    def even_modules(self) -> FanControlBuilder:
        """
        Select modules where (row + col) is even (checkerboard pattern).

        This creates a checkerboard selection starting with (0,0).
        Modules at positions like (0,0), (0,2), (1,1), (2,0), (2,2) are selected.

        Returns:
            FanControlBuilder: The builder instance for method chaining.

        """
        self._selected_parity = 0
        return self

    def odd_modules(self) -> FanControlBuilder:
        """
        Select modules where (row + col) is odd (checkerboard pattern).

        This creates a checkerboard selection opposite to even_modules().
        Modules at positions like (0,1), (1,0), (1,2), (2,1) are selected.

        Returns:
            FanControlBuilder: The builder instance for method chaining.

        """
        self._selected_parity = 1
        return self

    def downstream(self) -> FanControlBuilder:
        """
        Select the downstream layer (layer 0) for multi-layer modules (e.g. 0816).

        Note: This is used when you want to control the downstream layer only.
        If not set, all layers will be affected.
        Modules with only 1 layer will not be affected by this setting.

        Returns:
            FanControlBuilder: The builder instance for method chaining.

        """
        self._selected_layer = 0
        return self

    def upstream(self) -> FanControlBuilder:
        """
        Select the upstream layer (layer 1) for multi-layer modules (e.g. 0816).

        Note: This is used when you want to control the upstream layer only.
        If not set, all layers will be affected.
        Modules with only 1 layer will not be affected by this setting.

        Returns:
            FanControlBuilder: The builder instance for method chaining.

        """
        self._selected_layer = 1
        return self

    def fans(self, fans: list[int]) -> FanControlBuilder:
        """
        Select specific fans within selected modules.

        Args:
            fans (list[int]): List of 1-based fan indices to select.

        Notes:
            - Negative indices are NOT supported for fan selection.
            - You only need to use this method if you DON'T want to select all fans in the module.
            - If a module has fewer fans than specified, only the valid fans will be affected.
            - single fan modules will ignore and always be selected.

            - Fan indices start at 1 for the first fan in a module:
            ┌─────┬─────┬─────┐
            │  1  │  2  │  3  │
            ├─────┼─────┼─────┤
            │  4  │  5  │  6  │
            ├─────┼─────┼─────┤
            │  7  │  8  │  9  │
            └─────┴─────┴─────┘

        """
        if not fans:
            raise ValueError("fans list cannot be empty")
        if any(f == 0 for f in fans):
            raise ValueError("fans list cannot contain zero (use 1-based indexing)")
        if any(f < 0 for f in fans):
            raise ValueError("fans list cannot contain negative indices")
        self._selected_fans = [f - 1 for f in fans]
        return self

    def set_intensity(self, percent: float | list[float] | list[list[float]]) -> FanControlBuilder:
        """
        Set the intensity for the currently selected fans.

        Args:
            percent (float | list[float] | list[list[float]]): Intensity percentage(s) to set (0.0-100.0).
                - Single float: applies to all selected fans.
                - 1D list: spreads horizontally or vertically across modules (or fans if dimensions match).
                - 2D list matching module dimensions: one value per module.
                - 2D list matching fan grid dimensions: one value per fan (auto-detected).

        """
        rows = self._resolve_rows()
        cols = self._resolve_cols()

        # Get fan grid info for detection
        fan_grid_rows, fan_grid_cols, fan_rows_per_module, fan_cols_per_module, is_homogeneous = self._get_fan_grid_info(
            rows, cols
        )

        # Detect pattern level
        pattern_level = self._detect_pattern_level(percent, len(rows), len(cols), fan_grid_rows, fan_grid_cols)

        if pattern_level == "fan" and is_homogeneous:
            # Fan-level application
            fan_matrix = self._normalize_to_fan_matrix(percent, fan_grid_rows, fan_grid_cols)
            self._apply_fan_level_matrix(fan_matrix, rows, cols, fan_rows_per_module, fan_cols_per_module)
        else:
            # Module-level application (existing behavior)
            matrix = self._normalize_to_matrix(percent, len(rows), len(cols))

            for row_idx, row in enumerate(rows):
                for col_idx, col in enumerate(cols):
                    module = self.layout.get_module(row, col)
                    if module is None:
                        continue

                    # Check parity filter (checkerboard pattern)
                    if self._selected_parity is not None and (row + col) % 2 != self._selected_parity:
                        continue

                    pwm_value = matrix[row_idx][col_idx]
                    layers = self._resolve_layers(module)
                    fan_indices = self._resolve_fans(module)

                    for layer in layers:
                        for fan_idx in fan_indices:
                            self.layout.set_fan_pwm(row, col, layer, fan_idx, pwm_value)

        self._reset_selection()
        return self

    def set_intensity_func(self, func: Callable[[float, float], float]) -> FanControlBuilder:
        """
        Set intensity using a function of normalized position.

        Args:
            func: Callable (x, y) -> intensity where:
                - x: normalized column position (0.0 = left, 1.0 = right)
                - y: normalized row position (0.0 = top, 1.0 = bottom)
                - returns: intensity value (0.0-100.0), will be clamped

        Returns:
            FanControlBuilder for method chaining

        """
        rows = self._resolve_rows()
        cols = self._resolve_cols()

        # Calculate normalization factors (handle single row/col edge case)
        max_row = self.layout.nb_rows - 1 if self.layout.nb_rows > 1 else 1
        max_col = self.layout.nb_columns - 1 if self.layout.nb_columns > 1 else 1

        for row in rows:
            for col in cols:
                module = self.layout.get_module(row, col)
                if module is None:
                    continue

                # Check parity filter
                if self._selected_parity is not None and (row + col) % 2 != self._selected_parity:
                    continue

                # Normalize positions to 0-1
                x = col / max_col
                y = row / max_row

                # Call user function and clamp result
                intensity = max(0.0, min(100.0, func(x, y)))

                layers = self._resolve_layers(module)
                fan_indices = self._resolve_fans(module)

                for layer in layers:
                    for fan_idx in fan_indices:
                        self.layout.set_fan_pwm(row, col, layer, fan_idx, intensity)

        self._reset_selection()
        return self

    def apply(self, dry_run: bool = False) -> dict[str, dict[str, dict[str, list[int]]]] | None:
        """
        Apply the currently set fan intensity settings to the actual API

        Args:
            dry_run (bool): If True, will not send to API but return the payload instead

        Returns:
            The payload sent to the API, or None if an acceptable error occurred or in cooldown
                dict[str, dict[str, dict[str, list[int]]]] | None

        """
        if not self.layout.api_url:
            raise ValueError("API URL not set. Cannot apply changes to API.")

        # Check cooldown after layout mismatch
        if time.monotonic() - self._last_error_time < self.COOLDOWN_SECONDS:
            return None

        pwm_data: dict[str, dict[str, list[int]]] = {}

        for row, col in self.layout.modified_modules:
            module = self.layout.get_module(row, col)
            if not module:
                continue

            flat_index = self.layout.flat_index_map.get((row, col))
            if flat_index is None:
                continue

            flat_idex_str = str(flat_index)
            pwm_data[flat_idex_str] = {}

            for layer in range(module.num_layers):
                layer_pwms: list[int] = []
                for fan_idx in range(module.fans_per_layer):
                    pwm_pct = self.layout.get_fan_pwm(row, col, layer, fan_idx)
                    pwm_int = round(pwm_pct * 10)
                    layer_pwms.append(pwm_int)

                pwm_data[flat_idex_str][str(layer)] = layer_pwms

        payload = {"pwm_data": pwm_data}

        if dry_run:
            return payload

        url = f"{self.layout.api_url.rstrip('/')}/windcontrol/modules/pwms"
        response = requests.patch(url, json=payload, timeout=5)
        if response.status_code == 400:
            logger.warning("Layout mismatch - fan control skipped (layout may be updating)")
            self._last_error_time = time.monotonic()
            self.layout.modified_modules.clear()
            return None
        response.raise_for_status()
        self.layout.modified_modules.clear()

        return payload

    """
    Date: 09/01/2026
    Jonas Stirnemann

    Pattern Spreading Algorithm

    A bit complex, hard to explain concisely, but here's the gist:

    When calling set_intensity(percent), the system auto-detects whether to apply
    values at MODULE-level or FAN-level based on input dimensions.

    Detection rules (for 2x2 modules of 3x3 fans = 6x6 fan grid):
        Input exceeds module dimensions? -> FAN-level (spread/truncate to fan grid)
        Otherwise                        -> MODULE-level (spread to module grid)

    Examples:
        Input         Detection    Action
        -----         ---------    ------
        50.0          module       Same value for all fans
        [[a,b]]       module       2 cols <= 2 modules -> module-level, spread to 2x2
        [[a,b,c]]     fan          3 cols > 2 modules -> fan-level, spread to 6x6
        [[...],[...]] (2x2)        module       Matches module grid exactly
        [[...],[...],[...]] (3x6)  fan          Exceeds modules -> spread 3 rows to 6
        [[...]] (1x5)              fan          5 > 2 modules -> spread to 6x6

    Spreading algorithm:
        When input is smaller than target, values are repeated to fill:

        [A, B, C] spread to 6 cols:
            repeat_factor = 6 // 3 = 2
            remainder = 6 % 3 = 0
            Result: [A, A, B, B, C, C]

        [A, B, C] spread to 5 cols:
            repeat_factor = 5 // 3 = 1
            remainder = 5 % 3 = 2
            Result: [A, A, B, B, C]  (first 2 values get extra copy)

        When input is larger than target, values are truncated:
            [A, B, C, D, E] to 3 fans cols -> [A, B, C]

    """

    def _resolve_rows(self) -> list[int]:
        if self._selected_rows is not None:
            return self._selected_rows
        return list(range(self.layout.nb_rows))

    def _resolve_cols(self) -> list[int]:
        if self._selected_cols is not None:
            return self._selected_cols
        return list(range(self.layout.nb_columns))

    def _resolve_layers(self, module: ModuleConfig) -> list[int]:
        if self._selected_layer is not None:
            if module.num_layers > 1:
                return [self._selected_layer]
            return [0]
        return list(range(module.num_layers))

    def _resolve_fans(self, module: ModuleConfig) -> list[int]:
        if self._selected_fans is not None:
            if module.fans_per_layer > 1:
                valid_fans = [fan_idx for fan_idx in self._selected_fans if 0 <= fan_idx < module.fans_per_layer]
                return valid_fans if valid_fans else [0]
            return [0]
        return list(range(module.fans_per_layer))

    def _get_fan_grid_info(self, rows: list[int], cols: list[int]) -> tuple[int, int, int, int, bool]:
        """
        Calculate fan grid dimensions for the selected modules.

        Returns:
            tuple of (total_fan_rows, total_fan_cols, fan_rows_per_module,
                      fan_cols_per_module, is_homogeneous)

        """
        fan_rows_per_module: int | None = None
        fan_cols_per_module: int | None = None
        is_homogeneous = True

        for row in rows:
            for col in cols:
                module = self.layout.get_module(row, col)
                if module is None:
                    continue

                # Check parity filter
                if self._selected_parity is not None and (row + col) % 2 != self._selected_parity:
                    continue

                current_fan_rows = module.fan_grid_rows
                current_fan_cols = module.fan_grid_cols

                if fan_rows_per_module is None:
                    fan_rows_per_module = current_fan_rows
                    fan_cols_per_module = current_fan_cols
                elif fan_rows_per_module != current_fan_rows or fan_cols_per_module != current_fan_cols:
                    is_homogeneous = False

        # Default to 1x1 if no modules found
        fan_rows_per_module = fan_rows_per_module or 1
        fan_cols_per_module = fan_cols_per_module or 1

        total_fan_rows = len(rows) * fan_rows_per_module
        total_fan_cols = len(cols) * fan_cols_per_module

        return (total_fan_rows, total_fan_cols, fan_rows_per_module, fan_cols_per_module, is_homogeneous)

    def _detect_pattern_level(
        self,
        percent: float | list[float] | list[list[float]],
        nb_rows: int,
        nb_columns: int,
        fan_grid_rows: int,
        fan_grid_cols: int,
    ) -> str:
        """
        Detect whether the input should be applied at module-level or fan-level.

        Returns:
            'module' or 'fan'

        """
        # Scalar - always module level
        if isinstance(percent, (int, float)):
            return "module"

        # 2D matrix
        if len(percent) > 0 and isinstance(percent[0], list):
            matrix_rows = len(percent)
            matrix_cols = len(percent[0]) if percent[0] else 0

            # If input exceeds module dimensions, it must be fan-level intent
            # (e.g., 5 cols for 2 modules means fan-level, not module-level)
            cols_exceed_modules = matrix_cols > nb_columns and fan_grid_cols != nb_columns
            rows_exceed_modules = matrix_rows > nb_rows and fan_grid_rows != nb_rows

            if cols_exceed_modules or rows_exceed_modules:
                return "fan"

            # Default to module-level
            return "module"

        # 1D list - if length exceeds module count, treat as fan-level
        values_1d: list[float] = percent  # type: ignore[assignment]
        list_len = len(values_1d)

        # If list is longer than module columns (horizontal), use fan-level
        if list_len > nb_columns and fan_grid_cols != nb_columns:
            return "fan"

        # If list is longer than module rows (vertical context), use fan-level
        if list_len > nb_rows and fan_grid_rows != nb_rows:
            return "fan"

        # Default to module-level
        return "module"

    def _normalize_to_fan_matrix(
        self,
        percent: float | list[float] | list[list[float]],
        fan_grid_rows: int,
        fan_grid_cols: int,
    ) -> list[list[float]]:
        """Normalize input to a fan-level matrix."""
        # Handle scalar value
        if isinstance(percent, (int, float)):
            capped_value = max(0.0, min(100.0, float(percent)))
            return [[capped_value] * fan_grid_cols for _ in range(fan_grid_rows)]

        # Handle 2D matrix
        if len(percent) > 0 and isinstance(percent[0], list):
            matrix_2d: list[list[float]] = percent  # type: ignore[assignment]

            # Spread rows if needed
            if len(matrix_2d) < fan_grid_rows:
                matrix_2d = self._spread_rows(matrix_2d, fan_grid_rows)
            elif len(matrix_2d) > fan_grid_rows:
                matrix_2d = matrix_2d[:fan_grid_rows]

            # Spread columns for each row
            result: list[list[float]] = []
            for row_vals in matrix_2d:
                spread_cols = self._spread_values(list(row_vals), fan_grid_cols)
                capped_cols = [max(0.0, min(100.0, val)) for val in spread_cols]
                result.append(capped_cols)
            return result

        # Handle 1D list - spread based on selection context
        values_1d: list[float] = percent  # type: ignore[assignment]
        if self._selected_cols is not None and self._selected_rows is None:
            # Vertical spread
            spread_vertical = self._spread_values(values_1d, fan_grid_rows)
            capped_vertical = [max(0.0, min(100.0, val)) for val in spread_vertical]
            return [[val] * fan_grid_cols for val in capped_vertical]

        # Default: horizontal spread
        spread_horizontal = self._spread_values(values_1d, fan_grid_cols)
        capped_horizontal = [max(0.0, min(100.0, val)) for val in spread_horizontal]
        return [capped_horizontal[:] for _ in range(fan_grid_rows)]

    def _spread_rows(self, matrix: list[list[float]], target_rows: int) -> list[list[float]]:
        """Spread/repeat rows to match target row count."""
        if len(matrix) == 0:
            return [[0.0] for _ in range(target_rows)]
        if len(matrix) == 1:
            return [matrix[0][:] for _ in range(target_rows)]
        if len(matrix) >= target_rows:
            return matrix[:target_rows]

        repeat_factor = target_rows // len(matrix)
        remainder = target_rows % len(matrix)
        expanded: list[list[float]] = []
        for i, row in enumerate(matrix):
            count = repeat_factor + (1 if i < remainder else 0)
            expanded.extend([row[:] for _ in range(count)])
        return expanded

    def _apply_fan_level_matrix(
        self,
        fan_matrix: list[list[float]],
        rows: list[int],
        cols: list[int],
        fan_rows_per_module: int,
        fan_cols_per_module: int,
    ) -> None:
        """
        Apply a fan-level matrix to the selected modules.

        The fan_matrix dimensions should be:
        (len(rows) * fan_rows_per_module) x (len(cols) * fan_cols_per_module)

        """
        for row_idx, row in enumerate(rows):
            for col_idx, col in enumerate(cols):
                module = self.layout.get_module(row, col)
                if module is None:
                    continue

                # Check parity filter
                if self._selected_parity is not None and (row + col) % 2 != self._selected_parity:
                    continue

                layers = self._resolve_layers(module)

                # Calculate the starting position in the fan matrix for this module
                fan_row_start = row_idx * fan_rows_per_module
                fan_col_start = col_idx * fan_cols_per_module

                # Extract the sub-matrix for this module and apply to fans
                for local_fan_row in range(fan_rows_per_module):
                    for local_fan_col in range(fan_cols_per_module):
                        # Calculate fan index from grid position
                        fan_idx = local_fan_row * fan_cols_per_module + local_fan_col

                        # Skip if this fan is not in the selected fans
                        if self._selected_fans is not None and fan_idx not in self._selected_fans:
                            continue

                        # Skip if fan_idx exceeds module's actual fan count
                        if fan_idx >= module.fans_per_layer:
                            continue

                        # Get value from fan matrix
                        matrix_row = fan_row_start + local_fan_row
                        matrix_col = fan_col_start + local_fan_col

                        if matrix_row < len(fan_matrix) and matrix_col < len(fan_matrix[matrix_row]):
                            pwm_value = max(0.0, min(100.0, fan_matrix[matrix_row][matrix_col]))
                        else:
                            pwm_value = 0.0

                        for layer in layers:
                            self.layout.set_fan_pwm(row, col, layer, fan_idx, pwm_value)

    def _normalize_to_matrix(
        self,
        percent: float | list[float] | list[list[float]],
        nb_rows: int,
        nb_columns: int,
    ) -> list[list[float]]:
        # ! Handle scalar value
        if isinstance(percent, (int, float)):
            capped_value = max(0.0, min(100.0, float(percent)))
            return [[capped_value] * nb_columns for _ in range(nb_rows)]

        # ! Handle 2D matrix
        if len(percent) > 0 and isinstance(percent[0], list):
            matrix_2d: float | list[float] | list[list[float]] = percent  # type: ignore[assignment]
            # Spread rows to match nb_rows
            spread_row_count = len(matrix_2d)
            if spread_row_count != nb_rows:
                # We need to spread/repeat rows
                if spread_row_count == 0:
                    matrix_2d = [[0.0] * nb_columns for _ in range(nb_rows)]
                elif spread_row_count == 1:
                    matrix_2d = [matrix_2d[0][:] for _ in range(nb_rows)]
                elif spread_row_count < nb_rows:
                    repeat_factor = nb_rows // spread_row_count
                    remainder = nb_rows % spread_row_count
                    expanded: list[list[float]] = []
                    for i, row in enumerate(matrix_2d):
                        count = repeat_factor + (1 if i < remainder else 0)
                        expanded.extend([row[:] for _ in range(count)])
                    matrix_2d = expanded
                else:
                    matrix_2d = matrix_2d[:nb_rows]

            # Now spread columns for each row
            result: list[list[float]] = []
            for row_vals in matrix_2d:
                spread_cols = self._spread_values(row_vals, nb_columns)
                capped_cols = [max(0.0, min(100.0, val)) for val in spread_cols]
                result.append(capped_cols)
            return result

        # ! Handle 1D list
        values_1d: list[float] = percent  # type: ignore[assignment]
        if self._selected_rows is not None and self._selected_cols is None:
            spread_horizontal = self._spread_values(values_1d, nb_columns)
            capped_horizontal = [max(0.0, min(100.0, val)) for val in spread_horizontal]
            return [capped_horizontal[:] for _ in range(nb_rows)]
        if self._selected_cols is not None and self._selected_rows is None:
            spread_vertical = self._spread_values(values_1d, nb_rows)
            capped_vertical = [max(0.0, min(100.0, val)) for val in spread_vertical]
            return [[val] * nb_columns for val in capped_vertical]
        spread_horizontal = self._spread_values(values_1d, nb_columns)
        capped_horizontal = [max(0.0, min(100.0, val)) for val in spread_horizontal]
        return [capped_horizontal[:] for _ in range(nb_rows)]

    def _spread_values(self, values: list[float], target_size: int) -> list[float]:
        if len(values) == 0:
            return [0.0] * target_size
        if len(values) == 1:
            return [values[0]] * target_size
        if len(values) >= target_size:
            return values[:target_size]

        repeat_factor = target_size // len(values)
        remainder = target_size % len(values)

        result: list[float] = []
        for i, val in enumerate(values):
            count = repeat_factor + (1 if i < remainder else 0)
            result.extend([val] * count)

        return result
